/*
  matmul_jia.c: Matrix Multiplication (Inner Product) using the JIAJIA SWDSM.
  (K) 2002 Thomas Repantis <darkzero@otenet.gr>
*/

/* 
   This program calculates the inner product of two matrices using the 
   JIAJIA SWDSM.  Based on the documentation provided with JIAJIA.
*/

#include <stdio.h>
#include <jia.h>      /* To use JIAJIA API calls. */
#include <stdlib.h>   /* To use random(). */

#define N 1024	/* The size of the matrix. */
double (*a)[N], (*b)[N], (*c)[N];

void matr_init()
{
  int i,j;
  double swap;
  
  if (jiapid==0) /* Master's jobs. */ 
    {
      /* Fill the matrices to be multiplied with random numbers, 
	 which have a decimal part. */
      
      for (i=0;i<N;i++) 
	for (j=0;j<N;j++){
	  a[i][j]=( (double)random()/(double)RAND_MAX );
	  b[i][j]=( (double)random()/(double)RAND_MAX );
	}
    
      /* Find the transpose of matrix b to optimize the multiplication
	 by using locality. */
      for (i=0;i<N;i++)
	for (j=i+1;j<N;j++){
	  swap=b[i][j];
	  b[i][j]=b[j][i];
	  b[j][i]=swap;
	}
      
    }
}

void matr_multiply()
{
  int i,j,k;
  int start, end;
  double temp;
  
  if( N%jiahosts == 0 )
    start=(N/jiahosts)*jiapid;
  else
    start=( (N/jiahosts)+1 )*jiapid;
  if( N%jiahosts == 0 )
    end=start+(N/jiahosts);
  else
    end=start+(N/jiahosts)+1;
  
  for (j=0;j<N;j++){ 
    for (i=start;i<end;i++){
      temp=0.0;
      for (k=0;k<N;k++)
	/* Since we have found the transpose of b we can compute 
	   temp+=a[i][k]*b[j][k], instead of temp+=a[i][k]*b[k][j], 
	   improving the performance dramatically (~5 times faster). */
	temp+=a[i][k]*b[j][k];
      c[i][j]=temp;
      /* Produce some output (so that the user sees the progress), 
	 but do not show all of the product matrix elements (since 
	 they are too many), only the first of each column. */
      /* When time is measured, printf() is commented out. */
//      if (i==start)
//	printf("c[%d][%d]=%f\n",i,j,c[i][j]);
    }
  }
}


main(int argc,char **argv)
{
  float t1,t2;
  
  if (jiapid==0) 
    printf("\n\nmatmul_jia.c: Matrix Multiplication (Inner Product) using the JIAJIA SWDSM.\n(K) 2002 Thomas Repantis <darkzero@otenet.gr>\n\n");
  
  /* Initialize JIAJIA, create multiple processes. */
  jia_init(argc,argv);
  
  /* In order to reduce miss rate we allocate the matrices on the 
     processors uniformly.
     Allocate shared memory for each array:  
     N*N*sizeof(double) bytes allocated cyclically across all hosts, 
     each time (N*N*sizeof(double))/NumberOfHosts, starting the 
     allocation from host 0. */
  a=(double (*)[N])jia_alloc3(N*N*sizeof(double),(N*N*sizeof(double))/jiahosts,0);
  b=(double (*)[N])jia_alloc3(N*N*sizeof(double),(N*N*sizeof(double))/jiahosts,0);
  c=(double (*)[N])jia_alloc3(N*N*sizeof(double),(N*N*sizeof(double))/jiahosts,0);
  
  /* Perform a global barrier, to ensure that memory allocation is finished 
     on all hosts. */
  jia_barrier();
  
  /* Initialize the matrices to be multiplied. */
  matr_init();
  
  /* Perform a global barrier, to ensure that the matrix initialization is 
     finished. */
  jia_barrier();
  
  /* Start collecting statistics. */
  jia_startstat();
  
  /* Note the time the multiplication started. */
  t1=jia_clock();
  
  /* Start the worker program, multiple processes work on the shared data. */
  matr_multiply();
  
  /* Perform a global barrier, to ensure that the matrix multiplication is 
     finished on all hosts. */
  jia_barrier();
  
  /* Note the time the multiplication finished. */
  t2=jia_clock();
  
  if (jiapid==0) 
    printf("Total time needed for matrix multiplication:  %10.2f seconds\n", t2-t1);
  
  /* Exit JIAJIA. */
  jia_exit();
} 



