/*
  matmul_seq.c: Matrix Multiplication (Inner Product) sequential.
  (K) 2002 Thomas Repantis <darkzero@otenet.gr>
*/

/* 
   This program calculates the inner product of two matrices sequentially. 
   Based on the documentation provided with JIAJIA.
*/

#include <stdio.h>
#include <stdlib.h>   /* To use random(). */
#include <sys/time.h> /* To use gettimeofday(). */

#define MICROS_PER_SECOND 1000000.0
#define N 1024	/* The size of the matrix. */
double a[N][N], b[N][N], c[N][N];

void matr_init()
{
  int i,j;
  double swap;
  
  /* Fill the matrices to be multiplied with random numbers, 
     which have a decimal part. */
  for (i=0;i<N;i++) 
    for (j=0;j<N;j++){
      a[i][j]=( (double)random()/(double)RAND_MAX );
      b[i][j]=( (double)random()/(double)RAND_MAX );
    }
  
  /* Find the transpose of matrix b to optimize the multiplication
     by using locality. */
  for (i=0;i<N;i++)
    for (j=i+1;j<N;j++){
      swap=b[i][j];
      b[i][j]=b[j][i];
      b[j][i]=swap;
    }
}

void matr_multiply()
{
  int i,j,k;
  int start, end;
  double temp;
  
  start=0;
  end=N;
  
  for (j=0;j<N;j++){ 
    for (i=start;i<end;i++){
      temp=0.0;
      for (k=0;k<N;k++)
	/* Since we have found the transpose of b we can compute 
	   temp+=a[i][k]*b[j][k], instead of temp+=a[i][k]*b[k][j], 
	   improving the performance dramatically (~5 times faster). */
	temp+=a[i][k]*b[j][k];
      c[i][j]=temp;
      /* Produce some output (so that the user sees the progress), 
	 but do not show all of the product matrix elements (since 
	 they are too many), only the first of each column. */
      /* When time is measured, printf() is commented out. */
//      if (i==start)
//	printf("c[%d][%d]=%f\n",i,j,c[i][j]);
    }
  }
}


main(int argc,char **argv)
{
  float time_elapsed;
  struct timeval val1, val2;
  
  printf("\n\nmatmul_seq.c: Matrix Multiplication (Inner Product) sequential.\n(K) 2002 Thomas Repantis <darkzero@otenet.gr>\n\n");
  
  /* Initialize the matrices to be multiplied. */
  matr_init();
  
  /* Note the time the multiplication started. */
  gettimeofday(&val1,NULL);
  
  /* Start the worker program, one process works on the data sequentially. */
  matr_multiply();
  
  /* Note the time the multiplication finished. */
  gettimeofday(&val2,NULL);
  
  /* Calculate the seconds part of the time elapsed. */
  time_elapsed = (val2.tv_sec - val1.tv_sec)*MICROS_PER_SECOND;
  
  /* Calculate the microseconds part of the time elapsed. */
  time_elapsed += (val2.tv_usec - val1.tv_usec);
  
  /* Convert the time elapsed in seconds. */
  time_elapsed = time_elapsed  / MICROS_PER_SECOND;
  
  printf("Total time needed for matrix multiplication:  %10.2f seconds\n", time_elapsed);
  
} 



